==> fluidpythran-0.1.1.tar.gz: fluidpythran-0.1.1/fluidpythran/aheadoftime.py <==
"""User runtime API for the ahead-of-time compilation
=====================================================

User API
--------

.. autofunction:: pythran_def

.. autofunction:: make_signature

.. autoclass:: FluidPythran
   :members:
   :private-members:

Internal API
------------

.. autofunction:: _get_fluidpythran_calling_module

.. autoclass:: CheckCompiling
   :members:
   :private-members:

"""

import inspect
import time
from pathlib import Path
import subprocess

from .util import (
    get_module_name,
    has_to_pythranize_at_import,
    import_from_path,
    has_to_build,
    modification_date,
)

from .pythranizer import compile_pythran_file, name_ext_from_path_pythran

from .annotation import (
    make_signature_from_template_variables,
    make_signatures_from_typehinted_func,
)

from .log import logger

is_transpiling = False
modules = {}


def _get_fluidpythran_calling_module(index_frame: int = 2):
    """Get the FluidPythran instance corresponding to the calling module

    Parameters
    ----------

    index_frame : int

      Index (in :code:`inspect.stack()`) of the frame to be selected

    """

    try:
        frame = inspect.stack()[index_frame]
    except IndexError:
        print("index_frame", index_frame)
        print([frame[1] for frame in inspect.stack()])
        raise

    module_name = get_module_name(frame)

    if module_name in modules:
        fp = modules[module_name]
        if (
            fp._created_while_transpiling != is_transpiling
            or fp._pythranize_at_import_at_creation
            != has_to_pythranize_at_import()
            or hasattr(fp, "path_mod")
            and fp.path_mod.exists()
            and has_to_build(fp.path_pythran, fp.path_mod)
        ):
            fp = FluidPythran(frame=frame, reuse=False)
    else:
        fp = FluidPythran(frame=frame, reuse=False)

    return fp


def pythran_def(func):
    """Decorator to declare that a pythranized version of the function has to
    be used

    Parameters
    ----------

    func: a function

    """

    fp = _get_fluidpythran_calling_module()
    return fp.pythran_def(func)


def make_signature(func, **kwargs):
    """Create signature for a function with values for the template types

    Parameters
    ----------

    func: a function

    kwargs : dict

      The template types and their value

    """
    if not is_transpiling:
        return

    fp = _get_fluidpythran_calling_module()
    fp.make_signature(func, **kwargs)


class CheckCompiling:
    """Check if the module is been compiled and replace the module and the function"""

    def __init__(self, fp, func):
        self.has_been_replaced = False
        self.fp = fp
        self.func = func

    def __call__(self, *args, **kwargs):

        if self.has_been_replaced:
            return self.func(*args, **kwargs)

        fp = self.fp
        if fp.is_compiling and fp.process.poll() is not None:
            fp.is_compiling = False
            time.sleep(0.1)
            fp.module_pythran = import_from_path(
                fp.path_extension, fp.module_pythran.__name__
            )
            assert hasattr(self.fp.module_pythran, "__pythran__")
            fp.is_compiled = True

        if not fp.is_compiling:
            self.func = getattr(fp.module_pythran, self.func.__name__)
            self.has_been_replaced = True

        return self.func(*args, **kwargs)


class FluidPythran:
    """
    Representation of a module using ahead-of-time fluidpythran commands

    Parameters
    ----------

    use_pythran : bool (optional, default True)

      If False, don't use the pythranized versions at run time

    frame : int (optional)

      (Internal) Index (in :code:`inspect.stack()`) of the frame to be selected

    reuse : bool (optional, default True)

      (Internal) If True, do not recreate an instance.

    """

    def __init__(self, use_pythran=True, frame=None, reuse=True):

        if frame is None:
            frame = inspect.stack()[1]

        self.module_name = module_name = get_module_name(frame)

        self._pythranize_at_import_at_creation = has_to_pythranize_at_import()

        if reuse and module_name in modules:
            fp = modules[module_name]
            for key, value in fp.__dict__.items():
                self.__dict__[key] = value
            return

        self.names_template_variables = {}

        self._created_while_transpiling = is_transpiling

        if is_transpiling:
            self.functions = {}
            self.signatures_func = {}
            modules[module_name] = self
            self.is_transpiled = False
            return

        if not use_pythran:
            self.is_transpiled = False
            return

        self.is_compiling = False

        if "." in module_name:
            package, module_short_name = module_name.rsplit(".", 1)
            module_pythran_name = package + "."
        else:
            module_short_name = module_name
            module_pythran_name = ""

        module_pythran_name += "__pythran__._" + module_short_name

        self.path_mod = path_mod = Path(frame.filename)
        self.path_pythran = path_pythran = (
            path_mod.parent / "__pythran__" / ("_" + module_short_name + ".py")
        )

        path_ext = None

        if has_to_pythranize_at_import() and path_mod.exists():
            if has_to_build(path_pythran, path_mod):
                if path_pythran.exists():
                    time_pythran = modification_date(path_pythran)
                else:
                    time_pythran = 0
                print(f"Running fluidpythran on file {path_mod}... ", end="")
                # better to do this in another process because the file is already run...
                returncode = subprocess.call(
                    ["fluidpythran", "-np", str(path_mod)]
                )
                if returncode != 0:
                    raise RuntimeError(
                        "fluidpythran does not manage to produce the Pythran "
                        f"file for {path_mod}"
                    )

                print("Done!")

                path_ext = path_pythran.with_name(
                    name_ext_from_path_pythran(path_pythran)
                )

                time_pythran_after = modification_date(path_pythran)
                # We have to touch the files to signal that they are up-to-date
                if time_pythran_after == time_pythran:
                    if not has_to_build(path_ext, path_pythran):
                        path_pythran.touch()
                        if path_ext.exists():
                            path_ext.touch()
                    else:
                        path_pythran.touch()

        path_ext = path_ext or path_pythran.with_name(
            name_ext_from_path_pythran(path_pythran)
        )

        self.path_extension = path_ext

        if (
            has_to_pythranize_at_import()
            and path_mod.exists()
            and not self.path_extension.exists()
        ):
            print("Launching Pythran to compile a new extension...")
            self.process = compile_pythran_file(
                path_pythran, name_ext_file=self.path_extension.name
            )
            self.is_compiling = True
            self.is_compiled = False

        self.is_transpiled = True
        if path_ext.exists() and not self.is_compiling:
            self.module_pythran = import_from_path(
                self.path_extension, module_pythran_name
            )
        elif path_pythran.exists():
            self.module_pythran = import_from_path(
                path_pythran, module_pythran_name
            )
        else:
            self.is_transpiled = False
            self.is_compiled = False

        if self.is_transpiled:
            self.is_compiled = hasattr(self.module_pythran, "__pythran__")
            if self.is_compiled:
                module = inspect.getmodule(frame[0])
                module.__pythran__ = self.module_pythran.__pythran__
                module.__fluidpythran__ = self.module_pythran.__fluidpythran__

            if hasattr(self.module_pythran, "arguments_blocks"):
                self.arguments_blocks = getattr(
                    self.module_pythran, "arguments_blocks"
                )

        modules[module_name] = self

    def pythran_def(self, func):
        """Decorator used for functions

        Parameters
        ----------

        func: a function

        """

        if is_transpiling:
            self.functions[func.__name__] = func
            return func

        if not self.is_transpiled:
            return func

        try:
            func_tmp = getattr(self.module_pythran, func.__name__)
        except AttributeError:
            logger.warning("Pythran file does not seem to be up-to-date.")
            func_tmp = func

        if self.is_compiling:
            return CheckCompiling(self, func_tmp)

        return func_tmp

    def make_signature(self, func, _signature=None, **kwargs):
        """Create signature for a function with values for the template types

        Parameters
        ----------

        func: a function

        kwargs : dict

          The template types and their value

        """
        signature = make_signature_from_template_variables(
            func, _signature=_signature, **kwargs
        )

        if func.__name__ not in self.signatures_func:
            self.signatures_func[func.__name__] = []

        self.signatures_func[func.__name__].append(signature)

    def use_pythranized_block(self, name):
        """Use the pythranized version of a code block

        Parameters
        ----------

        name : str

          The name of the block.

        """
        if not self.is_transpiled:
            raise ValueError(
                "`use_pythranized_block` has to be used protected "
                "by `if fp.is_transpiled`"
            )

        if self.is_compiling and self.process.poll() is not None:
            self.is_compiling = False
            time.sleep(0.1)
            self.module_pythran = import_from_path(
                self.path_extension, self.module_pythran.__name__
            )
            assert hasattr(self.module_pythran, "__pythran__")
            self.is_compiled = True

        func = getattr(self.module_pythran, name)
        argument_names = self.arguments_blocks[name]

        frame = inspect.currentframe()
        try:
            locals_caller = frame.f_back.f_locals
        finally:
            del frame

        arguments = [locals_caller[name] for name in argument_names]
        return func(*arguments)

    def _make_signatures_from_annotations(self):
        """Make the signatures from annotations if it is possible

        Useful when there are only "not templated" types.

        """
        for func in self.functions.values():
            signatures = make_signatures_from_typehinted_func(func)
            if func.__name__ not in self.signatures_func:
                self.signatures_func[func.__name__] = []
            self.signatures_func[func.__name__].extend(signatures)

==> fluidpythran-0.1.4.tar.gz: fluidpythran-0.1.4/fluidpythran/aheadoftime.py <==
"""User runtime API for the ahead-of-time compilation
=====================================================

User API
--------

.. autofunction:: boost

.. autofunction:: pythran_def

.. autofunction:: make_signature

.. autoclass:: FluidPythran
   :members:
   :private-members:

Internal API
------------

.. autofunction:: _get_fluidpythran_calling_module

.. autoclass:: CheckCompiling
   :members:
   :private-members:

"""

import inspect
import time
import subprocess
import os
import functools
import sys

from .util import (
    get_module_name,
    has_to_pythranize_at_import,
    import_from_path,
    has_to_build,
    modification_date,
)

from .pythranizer import (
    compile_pythran_file,
    name_ext_from_path_pythran,
    ext_suffix,
)

from .annotation import (
    make_signature_from_template_variables,
    make_signatures_from_typehinted_func,
)

from .log import logger
from . import mpi
from .mpi import Path

if mpi.nb_proc == 1:
    mpi.has_to_build = has_to_build
    mpi.modification_date = modification_date

is_transpiling = False
modules = {}


def _get_fluidpythran_calling_module(index_frame: int = 2):
    """Get the FluidPythran instance corresponding to the calling module

    Parameters
    ----------

    index_frame : int

      Index (in :code:`inspect.stack()`) of the frame to be selected

    """

    try:
        frame = inspect.stack()[index_frame]
    except IndexError:
        print("index_frame", index_frame)
        print([frame[1] for frame in inspect.stack()])
        raise

    module_name = get_module_name(frame)

    if module_name in modules:
        fp = modules[module_name]
        if (
            fp.is_transpiling != is_transpiling
            or fp._pythranize_at_import_at_creation
            != has_to_pythranize_at_import()
            or hasattr(fp, "path_mod")
            and fp.path_mod.exists()
            and mpi.has_to_build(fp.path_pythran, fp.path_mod)
        ):
            fp = FluidPythran(frame=frame, reuse=False)
    else:
        fp = FluidPythran(frame=frame, reuse=False)

    return fp


def pythran_def(func):
    """Decorator to declare that a pythranized version of the function has to
    be used

    Parameters
    ----------

    func: a function

    """

    fp = _get_fluidpythran_calling_module()
    return fp.pythran_def(func)


def boost(obj):
    """Decorator to declare that an object can "use" pythran

    Parameters
    ----------

    obj: a function, a method or a class

    """

    fp = _get_fluidpythran_calling_module()
    return fp.boost(obj)


def make_signature(func, **kwargs):
    """Create signature for a function with values for the template types

    Parameters
    ----------

    func: a function

    kwargs : dict

      The template types and their value

    """
    if not is_transpiling:
        return

    fp = _get_fluidpythran_calling_module()
    fp.make_signature(func, **kwargs)


class CheckCompiling:
    """Check if the module is been compiled and replace the module and the function"""

    def __init__(self, fp, func):
        self.has_been_replaced = False
        self.fp = fp
        self.func = func

    def __call__(self, *args, **kwargs):

        if self.has_been_replaced:
            return self.func(*args, **kwargs)

        fp = self.fp
        if fp.is_compiling and not fp.process.is_alive():
            fp.is_compiling = False
            time.sleep(0.1)
            fp.module_pythran = import_from_path(
                fp.path_extension, fp.module_pythran.__name__
            )
            assert hasattr(self.fp.module_pythran, "__pythran__")
            fp.is_compiled = True

        if not fp.is_compiling:
            self.func = getattr(fp.module_pythran, self.func.__name__)
            self.has_been_replaced = True

        return self.func(*args, **kwargs)


class FluidPythran:
    """
    Representation of a module using ahead-of-time fluidpythran commands

    Parameters
    ----------

    use_pythran : bool (optional, default True)

      If False, don't use the pythranized versions at run time

    frame : int (optional)

      (Internal) Index (in :code:`inspect.stack()`) of the frame to be selected

    reuse : bool (optional, default True)

      (Internal) If True, do not recreate an instance.

    """

    def __init__(self, use_pythran=True, frame=None, reuse=True):

        if frame is None:
            frame = inspect.stack()[1]

        self.module_name = module_name = get_module_name(frame)

        self._pythranize_at_import_at_creation = has_to_pythranize_at_import()

        if reuse and module_name in modules:
            fp = modules[module_name]
            for key, value in fp.__dict__.items():
                self.__dict__[key] = value
            return

        self.names_template_variables = {}

        self.is_transpiling = is_transpiling

        if is_transpiling:
            self.functions = {}
            self.classes = {}
            self.signatures_func = {}
            modules[module_name] = self
            self.is_transpiled = False
            return

        if not use_pythran:
            self.is_transpiled = False
            return

        self.is_compiling = False

        if "." in module_name:
            package, module_short_name = module_name.rsplit(".", 1)
            module_pythran_name = package + "."
        else:
            module_short_name = module_name
            module_pythran_name = ""

        module_pythran_name += "__pythran__." + module_short_name

        self.path_mod = path_mod = Path(frame.filename)
        self.path_pythran = path_pythran = (
            path_mod.parent / "__pythran__" / (module_short_name + ".py")
        )

        path_ext = None

        if has_to_pythranize_at_import() and path_mod.exists():
            if mpi.has_to_build(path_pythran, path_mod):
                if path_pythran.exists():
                    time_pythran = mpi.modification_date(path_pythran)
                else:
                    time_pythran = 0

                returncode = None
                if mpi.rank == 0:
                    print(f"Running fluidpythran on file {path_mod}... ", end="")
                    # better to do this in another process because the file is already run...
                    os.environ["FLUIDPYTHRAN_NO_MPI"] = "1"
                    returncode = subprocess.call(
                        [
                            sys.executable,
                            "-m",
                            "fluidpythran.run",
                            "-np",
                            str(path_mod),
                        ]
                    )
                    del os.environ["FLUIDPYTHRAN_NO_MPI"]
                returncode = mpi.bcast(returncode)

                if returncode != 0:
                    raise RuntimeError(
                        "fluidpythran does not manage to produce the Pythran "
                        f"file for {path_mod}"
                    )

                if mpi.rank == 0:
                    print("Done!")

                path_ext = path_pythran.with_name(
                    name_ext_from_path_pythran(path_pythran)
                )

                time_pythran_after = mpi.modification_date(path_pythran)
                # We have to touch the files to signal that they are up-to-date
                if time_pythran_after == time_pythran and mpi.rank == 0:
                    if not has_to_build(path_ext, path_pythran):
                        path_pythran.touch()
                        if path_ext.exists():
                            path_ext.touch()
                    else:
                        path_pythran.touch()

        path_ext = path_ext or path_pythran.with_name(
            name_ext_from_path_pythran(path_pythran)
        )

        self.path_extension = path_ext

        if (
            has_to_pythranize_at_import()
            and path_mod.exists()
            and not self.path_extension.exists()
        ):
            if mpi.rank == 0:
                print("Launching Pythran to compile a new extension...")
            self.process = compile_pythran_file(
                path_pythran, name_ext_file=self.path_extension.name
            )
            self.is_compiling = True
            self.is_compiled = False

        self.is_transpiled = True

        if not path_ext.exists() and not self.is_compiling:
            path_ext_alt = path_pythran.with_suffix(ext_suffix)
            if path_ext_alt.exists():
                self.path_extension = path_ext = path_ext_alt

        self.reload_module_pythran(module_pythran_name)

        if self.is_transpiled:
            self.is_compiled = hasattr(self.module_pythran, "__pythran__")
            if self.is_compiled:
                module = inspect.getmodule(frame[0])
                # module can be None if (at least) it has been run with runpy
                if module is not None:
                    module.__pythran__ = self.module_pythran.__pythran__
                    module.__fluidpythran__ = self.module_pythran.__fluidpythran__

            if hasattr(self.module_pythran, "arguments_blocks"):
                self.arguments_blocks = getattr(
                    self.module_pythran, "arguments_blocks"
                )

        modules[module_name] = self

    def reload_module_pythran(self, module_pythran_name=None):
        if module_pythran_name is None:
            module_pythran_name = self.module_pythran.__name__
        if self.path_extension.exists() and not self.is_compiling:
            self.module_pythran = import_from_path(
                self.path_extension, module_pythran_name
            )
        elif self.path_pythran.exists():
            self.module_pythran = import_from_path(
                self.path_pythran, module_pythran_name
            )
        else:
            self.is_transpiled = False
            self.is_compiled = False

    def pythran_def(self, func):
        """Decorator used for functions

        Parameters
        ----------

        func: a function

        """

        signature = inspect.signature(func)
        try:
            is_method = next(iter(signature.parameters.keys())) == "self"
        except StopIteration:
            is_method = False

        if is_method:
            return self.pythran_def_method(func)

        if is_transpiling:
            self.functions[func.__name__] = func
            return func

        if not self.is_transpiled:
            return func

        if not hasattr(self.module_pythran, func.__name__):
            self.reload_module_pythran()

        try:
            func_tmp = getattr(self.module_pythran, func.__name__)
        except AttributeError:
            logger.warning("Pythran file does not seem to be up-to-date.")
            func_tmp = func

        if self.is_compiling:
            return functools.wraps(func)(CheckCompiling(self, func_tmp))

        return func_tmp

    def pythran_def_method(self, func):
        """Decorator used for methods

        Parameters
        ----------

        func: a function

        """
        if is_transpiling:
            func.__fluidpythran__ = "pythran_def_method"
            return func

        if not self.is_transpiled:
            return func

        return FluidPythranTemporaryMethod(func)

    def boost(self, obj):
        """Universal decorator for aot compilation"""
        if isinstance(obj, type):
            return self.pythran_class(obj)
        else:
            return self.pythran_def(obj)

    def pythran_class(self, cls: type):
        """Decorator used for classes

        Parameters
        ----------

        cls: a class

        """
        if is_transpiling:
            self.classes[cls.__name__] = cls
            return cls

        if not self.is_transpiled:
            return cls

        cls_name = cls.__name__

        for key, value in cls.__dict__.items():
            if not isinstance(value, FluidPythranTemporaryMethod):
                continue
            func = value.func
            func_name = func.__name__

            name_pythran_func = f"__for_method__{cls_name}__{func_name}"
            name_var_code_new_method = (
                f"__code_new_method__{cls_name}__{func_name}"
            )

            if not hasattr(self.module_pythran, name_pythran_func):
                self.reload_module_pythran()

            try:
                pythran_func = getattr(self.module_pythran, name_pythran_func)
                code_new_method = getattr(
                    self.module_pythran, name_var_code_new_method
                )
            except AttributeError:
                logger.warning("Pythran file does not seem to be up-to-date.")
                # setattr(cls, key, func)
            else:
                namespace = {"pythran_func": pythran_func}
                exec(code_new_method, namespace)
                setattr(cls, key, functools.wraps(func)(namespace["new_method"]))

        return cls

    def make_signature(self, func, _signature=None, **kwargs):
        """Create signature for a function with values for the template types

        Parameters
        ----------

        func: a function

        kwargs : dict

          The template types and their value

        """
        signature = make_signature_from_template_variables(
            func, _signature=_signature, **kwargs
        )

        if func.__name__ not in self.signatures_func:
            self.signatures_func[func.__name__] = []

        self.signatures_func[func.__name__].append(signature)

    def use_pythranized_block(self, name):
        """Use the pythranized version of a code block

        Parameters
        ----------

        name : str

          The name of the block.

        """
        if not self.is_transpiled:
            raise ValueError(
                "`use_pythranized_block` has to be used protected "
                "by `if fp.is_transpiled`"
            )

        if self.is_compiling and not self.process.is_alive():
            self.is_compiling = False
            time.sleep(0.1)
            self.module_pythran = import_from_path(
                self.path_extension, self.module_pythran.__name__
            )
            assert hasattr(self.module_pythran, "__pythran__")
            self.is_compiled = True

        func = getattr(self.module_pythran, name)
        argument_names = self.arguments_blocks[name]

        frame = inspect.currentframe()
        try:
            locals_caller = frame.f_back.f_locals
        finally:
            del frame

        arguments = [locals_caller[name] for name in argument_names]
        return func(*arguments)

    def _make_signatures_from_annotations(self):
        """Make the signatures from annotations if it is possible

        Useful when there are only "not templated" types.

        """
        for func in self.functions.values():
            signatures = make_signatures_from_typehinted_func(func)
            if func.__name__ not in self.signatures_func:
                self.signatures_func[func.__name__] = []
            self.signatures_func[func.__name__].extend(signatures)


class FluidPythranTemporaryMethod:
    def __init__(self, func):
        self.func = func

    def __call__(self, self_bis, *args, **kwargs):
        raise RuntimeError(
            "Did you forget to decorate a class using methods decorated "
            "with fluidpythran? Please decorate it with @boost."
        )
